"use strict";

let _               =   require('underscore'),
    AWS             =   require('aws-sdk'),
    sns             =   null,
    utilities       =   require('../utilities/index'),
    config          =   require('./config'),
    mailer          =   require('../mailer/index'),
    interfaces      =   require('../interfaces/index'),
    models          =   require ('mongoose').models,

    paramReplace    =   {
                            param1: (str, data) => str.replace (/param1/g, data),
                            param2: (str, data) => str.replace (/param2/g, data),
                            param3: (str, data) => str.replace (/param3/g, data),
                            param4: (str, data) => str.replace (/param4/g, data),
                            param5: (str, data) => str.replace (/param5/g, data)
                        },
    paramKeys       =   _.keys(paramReplace),
    test            =   'TEST',
    services        =   {};


let logErrorAndReturn =     function (e, result, callback) {
                                let err = null;
                                if (e) {
                                    err = new utilities.error.aws (e.code);
                                    if (err.code === 824 || err.code === 806) utilities.logger.warn(e);
                                    else {
                                        utilities.logger.error(err);
                                        utilities.logger.error(e);
                                    }
                                }
                                if (callback) return callback (err, result);
                            };

function getChatsUnseenCount (data, callback){

    let getArgs = {
            _person:    data.person._id,
            archived:   false,
            unseen:     true
        };

    models.Chat
        .where( getArgs )
        .count( function (e, count ) {
            if (e) { logger.error(e); return callback (new utilities.error.db (102)); }
            return callback (null, count);
        });
};

function getNotificationsUnseenCount (data, callback){

    let getArgs = {
            _person: data.person._id,
            unseen: true
        };

    models.Notification
        .where( getArgs )
        .count( function (e, count ) {
            if (e) {
                logger.error(e);
                return callback (new utilities.error.db (102));
            }
            return callback (null, count);
        });
};

function getNotificationsFollowUnseenCount (data, callback){

    let getArgs = {
            _person: data.person._id,
            unseen: true
        };

    return models.NotificationFollow.where(getArgs).count()
        .then(count => callback (null, count))
        .catch( e => {
            logger.error(e);
            return callback (new error.db (102));
        });
};


function getUnseenCounts (data, callback) {
    let error = null,
        results = {newChats: 0, newMsgs: 0, newFollowPosts: 0},
        doneFunc = _.after (3, () => callback (error, results) );

    getChatsUnseenCount (data, function (e, chatCount) {
        if (e) error = e;
        else results.newChats = chatCount;
        doneFunc();
    });

    getNotificationsUnseenCount (data, function (e, commentsCount) {
        if (e) error = e;
        else results.newMsgs = commentsCount;
        doneFunc();
    });

    getNotificationsFollowUnseenCount (data, function (e, notifyFollowCount) {
        if (e) error = e;
        else results.newFollowPosts = notifyFollowCount;
        doneFunc();
    });
}

function replaceParameters (data) {
    let message = data.message;
    if (data.msgParams) _.each ( _.pick (data.msgParams, paramKeys), (parameter, paramName) => { message = paramReplace [paramName] (message, parameter); }, this);
    return message;
}


function buildAWSSNSMessage (data, callback) {
    let counts =        { },
        badgeCount =    0,
        notification =  _.extend ( {}, config.notifications[data.id]),
        buildMessage =  function () {
                            notification.message = replaceParameters ({msgParams:data.msgParams, message:notification.message});

                            let message =   { default: notification.message },
                                apnsMsg =   { aps: {}, id: data.id },
                                gcmMsg =    { delay_while_idle: false, data: { id: data.id, message: notification.message } };

                            if (notification.silent) apnsMsg.aps = { "content-available": 1 };
                            else apnsMsg.aps = { alert: notification.message, badge: badgeCount, sound: notification.sound || 'bingbong.aiff' };

                            if ( data.custom ) {
                                _.extend ( apnsMsg, data.custom, counts );
                                _.extend ( gcmMsg.data, data.custom, counts );
                            }

                            message [config.apnsId] = JSON.stringify (apnsMsg);
                            message [config.gcmId] = JSON.stringify (gcmMsg);
                            return JSON.stringify(message);
                        };

    if (notification.silent) return callback (null, buildMessage() );
    else getUnseenCounts (data, function(e, results){
        if (!e) {
            badgeCount = results.newChats + results.newMsgs;
            counts = results;
        }
        return callback (null, buildMessage());
    });
};


function getProfile (person, callback) {
    if (!_.has(person, '_id') || !_.has(person, 'notifications') || !_.has(person, 'stopNotify') || !_.has(person, 'stopNotifyEmail') || !_.has(person, 'emailInvalid')) {
        models.Person.findOne({ _id: person._id })
            .lean().exec( (e, fullPerson) => {
                if (e) {
                    new utilities.error.db(100);
                    utilities.logger.error(e);
                    return callback ( null );
                } else return callback (fullPerson);
            });
    } else return callback (person);
}


//// Public --------------------------------------------------------------------------------------

module.exports.init = function (options)  {
    /*if (options && _.isObject(options)) _.extend(config, options);
    // Set up enabled services
    _.each (config.services, function (service, serviceName){
        if (service.enabled) {
            services[serviceName] = require('./bin/' + serviceName);
            services[serviceName].init (service);
        }

    });*/

    utilities.logger.info('framework/notify init');
    if (options) _.extend (config, options);
    if (config.enabled) {
        AWS.config.update (_.pick(config, 'accessKeyId', 'secretAccessKey', 'region'));
        sns = new AWS.SNS({region: config.region});
    }
};

module.exports.addDevice = function (data, callback) { //data = {deviceToken: String, _id: String, platform: String}
    utilities.logger.info('framework/notify addDevice');
    let params = { PlatformApplicationArn: config[ data.platform + 'PlatformApplicationArn' ], Token: data.deviceToken };
    // services.awssns.addDevice(params, callback);
    sns.createPlatformEndpoint ( params, (e, result) => logErrorAndReturn (e, result, callback) );
};

module.exports.removeDevice = function (data, callback) {
    utilities.logger.info('framework/notify removeDevice');
    if (!callback) callback = (e,result)=>{};
    let params = { EndpointArn: data.endpointArn };
    // services.awssns.removeDevice (params, callback );
    sns.deleteEndpoint ( params, (e, result) => logErrorAndReturn (e, result, callback) );

};
module.exports.disableDevice = function (data, callback) {
    utilities.logger.info('framework/notify disableDevice');
    let params = { EndpointArn: data.endpointArn, Attributes: { Enabled:'false' } };
    // services.awssns.setEndpointAttributes (params, callback);
    sns.setEndpointAttributes ( params, (e, result) => logErrorAndReturn (e, result, callback) );
};

module.exports.getDeviceAttributes = function (data, callback) {
    utilities.logger.info('framework/notify getDeviceAttributes');
    let params = { EndpointArn: data.endpointArn };
    // services.awssns.getEndpointAttributes (params, callback);
    sns.getEndpointAttributes ( params, (e, result) => logErrorAndReturn (e, result, callback) );
};

module.exports.setDeviceToken = function (data, callback) {
    utilities.logger.info('framework/notify setDeviceToken');

    let params = { EndpointArn: data.endpointArn, Attributes: { Enabled: 'true', Token: data.deviceToken } };
    // services.awssns.setEndpointAttributes (params, callback);
    sns.setEndpointAttributes ( params, (e, result) => logErrorAndReturn (e, result, callback) );
};

module.exports.enableDevice = function (data, callback) {
    utilities.logger.info('framework/notify enableDevice');
    let params = { EndpointArn: data.endpointArn, Attributes: { Enabled: 'true' } };
    // services.awssns.setEndpointAttributes (params, callback);
    sns.setEndpointAttributes ( params, (e, result) => logErrorAndReturn (e, result, callback) );
};

module.exports.disableDevice = function (data, callback) {
    utilities.logger.info('framework/notify disableDevice');

    let params = { EndpointArn: data.endpointArn, Attributes: { Enabled: 'false' } };
    // services.awssns.setEndpointAttributes (params, callback);
    sns.setEndpointAttributes ( params, (e, result) => logErrorAndReturn (e, result, callback) );
};

module.exports.subscribeToTopic = function (data, callback) {
    utilities.logger.info('framework/notify subscribeToTopic');
    let params = { Protocol: data.protocol, Endpoint: data.endpointArn, TopicArn: data.topicArn };
    sns.subscribe (params, (e, result) => logErrorAndReturn (e, result, callback) );
};

let unsubscribeFromTopic =  module.exports.unsubscribeFromTopic = function (data, callback) {
    utilities.logger.info('framework/notify unsubscribeFromTopic');
    let params = { SubscriptionArn: data.subscriptionArn };
    sns.unsubscribe(params, (e, result) => logErrorAndReturn (e, result, callback) );
};

/*module.exports.unsubscribe = function (data, callback) {
    utilities.logger.info('framework/notify subscribeToTopic');
    let params = { Protocol: data.protocol, Endpoint: data.endpointArn, TopicArn: data.topicArn };
    sns.subscribe (params, (e, result) => logErrorAndReturn (e, result, callback) );
};*/

if (process.env.testing === 'true') {
    module.exports.sendNotification = function (data, callback) { //data = {person: Object, id: String, badge: number, custom: Object, msgParams: Object}
        utilities.logger.info('framework/notify sendNotification');
        if (!callback) callback = () => {};
        // if (!services.awssns)              return callback (null, null);
        if (!data.person.notifications )        return callback (null, null);
        if (!data.person.notifications.length)  return callback (null, null);
        if (!config.notifications[data.id]) { utilities.logger.warn ('**** Notification type does not exist ***'); return callback (null, null); }

        //if ( data.person.settings && data.person.settings.stopNotify && data.person.settings.stopNotify.length && _.indexOf(data.person.settings.stopNotify, data.id) > -1 ) return utilities.logger.info ( data.person.username + ' has turned off ' + data.id + ' notification');
        if ( data.person.settings && data.person.settings.stopNotify && data.person.settings.stopNotify.length && data.person.settings.stopNotify.find (id => id === data.id) ) return utilities.logger.info ( data.person.username + ' has turned off ' + data.id + ' notification');

        buildAWSSNSMessage(data, function (e, message) {
            _.each (data.person.notifications, function(registration){
                // if (registration.targetArn) services.awssns.sendNotification ({ TargetArn: registration.targetArn, Message: message, MessageStructure: 'json' }, callback);
                if (registration.targetArn) sns.publish ( { TargetArn: registration.targetArn, Message: message, MessageStructure: 'json' }, (e, result) => logErrorAndReturn (e, result, callback) );
                else return callback(null, null);
            });
        });


    };
} else {
    module.exports.sendNotification = function (data) { //data = {person: Object, id: String, badge: number, custom: Object, msgParams: Object}
        utilities.logger.info('framework/notify sendNotification');
        if ( !config.notifications[data.id] ) return utilities.logger.warn ('**** Notification type does not exist ***');

        getProfile (data.person, person => {
            if (!person) {
                utilities.logger.info ( "Person not found" );
                utilities.logger.info ( JSON.stringify(data.person) );
                return new utilities.error.app(401);
            }

            if (person.emailInvalid === true) utilities.logger.info ( person.username + '\'s email address (' + person.email +') is invalid');
            else {
                if ( person.stopNotifyEmail && person.stopNotifyEmail.length && person.stopNotifyEmail.find (id => id === data.id || id === 'all' || id === "disabled")  )  utilities.logger.info ( person.username + ' has turned off ' + data.id + ' email notification');
                else if (data.mailParams) {

                    let mailerArgs = {
                        to:         person.email,
                        product:    'shufl',
                        locale:     'en',
                        person:     person,
                        mailType:   data.id,
                        msgParams:  data.msgParams,
                        mailParams: data.mailParams
                    };
                    mailer.send ( mailerArgs, result => {} );
                }
            }

            if ( !person.notifications )           return utilities.logger.info ('No notifications object found for user');
            if ( !person.notifications.length )    return utilities.logger.info ( person.username + ' does not have notifications enabled');

            if (data.id !== 'massMessage' && person.stopNotify && person.stopNotify.length && person.stopNotify.find(id => id === data.id || id === 'all') ) utilities.logger.info ( person.username + ' has turned off ' + data.id + ' notification');
            else {
                utilities.logger.info ( (person.username + ' has ' + person.notifications.length + ' devices registered') );

                let sendTheNotifications = function (e, message) {
                    let removeList = [],
                        newNotifications = [],
                        errorDone = _.after(person.notifications.length, function () {
                            if (removeList.length) {
                                newNotifications = _.reject (person.notifications,  notification => {
                                    let removeTarget = removeList.find(registration => registration.targetArn === notification.targetArn);
                                    if (removeTarget && removeTarget.topics && removeTarget.topics.global) unsubscribeFromTopic ( { subscriptionArn: removeTarget.topics.global } );
                                    return removeTarget;
                                });
                                utilities.logger.info ("Removing " + removeList.length + " SNS disabled or invalid endpoint")
                                models.Person
                                    .where({_id:person._id})
                                    .update( { notifications: newNotifications }, e => { if (e) {new utilities.error.db(100); utilities.logger.error(e);} });
                            }
                        });

                    _.each (person.notifications, function(registration){
                        if (registration.targetArn) {
                            utilities.logger.info ( 'Sending notification to @' + person.username  + ' using target: ' + registration.targetArn);
                            // services.awssns.sendNotification ({ TargetArn: registration.targetArn, Message: message, MessageStructure: 'json' }, function (e) {
                            sns.publish ( { TargetArn: registration.targetArn, Message: message, MessageStructure: 'json' }, e => {
                                if (e && ( e.code === 824 || e.code === 806 ) ) {
                                    removeList.push (registration);
                                    services.awssns.removeDevice ( { EndpointArn: registration.targetArn } );
                                }
                                errorDone();
                            });
                        } else {
                            utilities.logger.info('Weird. No target found.')
                            errorDone();
                        }
                    });
                };
                buildAWSSNSMessage( data, sendTheNotifications );
            }
        });
    };
}













/*function buildAWSSNSMessageOld (data, callback) {
    let badgeCount = 0,
        notification = _.extend ( {}, config.notifications[data.id]);

    getUnseenCounts (data, function(e, results){

        if (!e) badgeCount = results.chats + results.comments;
        if (data.msgParams) _.each ( _.pick (data.msgParams, 'param1', 'param2', 'param3', 'param4', 'param5'), function(parameter, paramName) { notification.message = paramReplace [paramName] (notification.message, parameter); }, this);

        let message = { default: notification.message },
            apnsMsg = {
                aps: {
                    alert: notification.message,
                    badge: badgeCount,
                    sound: notification.sound || 'bingbong.aiff',
                    // "content-available": 1
                },
                id: data.id
            },
            gcmMsg = {
                delay_while_idle: false,
                //collapse_key: Welcome,
                //time_to_live: 125,
                //dry_run: false,
                data: {
                    id: data.id,
                    message: notification.message,
                    //url: "http://www.amazon.com/"
                }
            };
            //mpnsMsg = "<?xml version=\"1.0\" encoding=\"utf-8\"?><wp:Notification xmlns:wp=\"WPNotification\"><wp:Tile><wp:Count>" + 0 + "</wp:Count><wp:Title>" + notification.message + "</wp:Title><wp:Custom1>shufl</wp:Custom1></wp:Tile></wp:Notification>";


        if (data.custom && _.isObject(data.custom)) {
            _.extend ( apnsMsg, data.custom, { newChats: results.chats, newMsgs: results.comments} );
            _.extend ( gcmMsg.data, data.custom, { newChats: results.chats, newMsgs: results.comments} );
            //_.extend ( apnsMsg, data.custom, { newChats: 0, newMsgs: 0 } );
            // _.extend ( gcmMsg.data, data.custom );
            //_.extend ( mpnsMsg, data.custom );

        }

        message [config.services.awssns.apnsId] = JSON.stringify (apnsMsg);
        message [config.services.awssns.gcmId] = JSON.stringify (gcmMsg);
        //message [config.awssns.mpnsId] = mpnsMsg;

        return callback (null, JSON.stringify(message));
    });
};*/